@@ -48,6 +48,7 @@
 
 HTMLTextClass html_text_class;
 static HTMLObjectClass *parent_class = NULL;
+static const PangoAttrClass html_pango_attr_font_size_klass;
 
 #define HT_CLASS(x) HTML_TEXT_CLASS (HTML_OBJECT (x)->klass)
 
@@ -57,7 +58,7 @@ static void         move_spell_errors       (GList *spell_errors, guint offset,
 static GList *      remove_spell_errors     (GList *spell_errors, guint offset, guint len);
 static void         remove_text_slaves      (HTMLObject *self);
 
-/* static void
+/* void
 debug_spell_errors (GList *se)
 {
 	for (;se;se = se->next)
@@ -150,6 +151,7 @@ copy (HTMLObject *s,
 	dest->select_start  = src->select_start;
 	dest->select_length = src->select_length;
 	dest->attr_list     = pango_attr_list_copy (src->attr_list);
+	dest->extra_attr_list = src->extra_attr_list ? pango_attr_list_copy (src->extra_attr_list) : NULL;
 
 	html_color_ref (dest->color);
 
@@ -232,7 +234,7 @@ cut_attr_list_filter (PangoAttribute *attr, gpointer data)
 }
 
 static void
-cut_attr_list (HTMLText *text, gint begin_index, gint end_index)
+cut_attr_list_list (PangoAttrList *attr_list, gint begin_index, gint end_index)
 {
 	PangoAttrList *removed;
 	PangoAttribute range;
@@ -240,11 +242,19 @@ cut_attr_list (HTMLText *text, gint begin_index, gint end_index)
 	range.start_index = begin_index;
 	range.end_index = end_index;
 
-	removed = pango_attr_list_filter (text->attr_list, cut_attr_list_filter, &range);
+	removed = pango_attr_list_filter (attr_list, cut_attr_list_filter, &range);
 	if (removed)
 		pango_attr_list_unref (removed);
 }
 
+static void
+cut_attr_list (HTMLText *text, gint begin_index, gint end_index)
+{
+	cut_attr_list_list (text->attr_list, begin_index, end_index);
+	if (text->extra_attr_list)
+		cut_attr_list_list (text->extra_attr_list, begin_index, end_index);
+}
+
 static void
 cut_links (HTMLText *text, gint start_offset, gint end_offset)
 {
@@ -347,6 +357,11 @@ html_text_op_cut_helper (HTMLText *text, HTMLEngine *e, GList *from, GList *to,
 			cut_links (rvt, 0, begin);
 		nt = g_strconcat (text->text, tail, NULL);
 		g_free (text->text);
+
+		rvt->spell_errors = remove_spell_errors (rvt->spell_errors, 0, begin);
+		rvt->spell_errors = remove_spell_errors (rvt->spell_errors, end, text->text_len - end);
+		move_spell_errors (rvt->spell_errors, begin, -begin);
+
 		text->text = nt;
 		text->text_len -= end - begin;
 		*len           += end - begin;
@@ -359,6 +374,7 @@ html_text_op_cut_helper (HTMLText *text, HTMLEngine *e, GList *from, GList *to,
 
 		text->spell_errors = remove_spell_errors (text->spell_errors, begin, end - begin);
 		move_spell_errors (text->spell_errors, end, - (end - begin));
+
 		html_text_convert_nbsp (text, TRUE);
 		pango_info_destroy (text);
 	} else {
@@ -441,11 +457,22 @@ object_merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, G
 		e->cursor->offset += t1->text_len;
 	}
 
+	/* printf ("--- before merge\n");
+	   debug_spell_errors (t1->spell_errors);
+	   printf ("---\n");
+	   debug_spell_errors (t2->spell_errors);
+	   printf ("---\n");
+	*/
 	move_spell_errors (t2->spell_errors, 0, t1->text_len);
 	t1->spell_errors = g_list_concat (t1->spell_errors, t2->spell_errors);
 	t2->spell_errors = NULL;
 
 	pango_attr_list_splice (t1->attr_list, t2->attr_list, t1->text_bytes, t2->text_bytes);
+	if (t2->extra_attr_list) {
+		if (!t1->extra_attr_list)
+			t1->extra_attr_list = pango_attr_list_new ();
+		pango_attr_list_splice (t1->extra_attr_list, t2->extra_attr_list, t1->text_bytes, t2->text_bytes);
+	}
 	merge_links (t1, t2);
 
 	to_free       = t1->text;
@@ -459,8 +486,8 @@ object_merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, G
 	pango_info_destroy (t2);
 
 	/* html_text_request_word_width (t1, e->painter); */
-	/* printf ("merged '%s'\n", t1->text); */
-	/* printf ("--- after merge\n");
+	/* printf ("merged '%s'\n", t1->text);
+	   printf ("--- after merge\n");
 	   debug_spell_errors (t1->spell_errors);
 	   printf ("---\n"); */
 
@@ -505,9 +532,19 @@ split_attrs (HTMLText *t1, HTMLText *t2, gint index)
 	delete = pango_attr_list_filter (t1->attr_list, split_attrs_filter_head, GINT_TO_POINTER (index));
 	if (delete)
 		pango_attr_list_unref (delete);
+	if (t1->extra_attr_list) {
+		delete = pango_attr_list_filter (t1->extra_attr_list, split_attrs_filter_head, GINT_TO_POINTER (index));
+		if (delete)
+			pango_attr_list_unref (delete);
+	}
 	delete = pango_attr_list_filter (t2->attr_list, split_attrs_filter_tail, GINT_TO_POINTER (index));
 	if (delete)
 		pango_attr_list_unref (delete);
+	if (t2->extra_attr_list) {
+		delete = pango_attr_list_filter (t2->extra_attr_list, split_attrs_filter_tail, GINT_TO_POINTER (index));
+		if (delete)
+			pango_attr_list_unref (delete);
+	}
 }
 
 static void
@@ -620,11 +657,12 @@ object_split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, g
 	move_spell_errors   (HTML_TEXT (dup)->spell_errors, 0, - HTML_TEXT (self)->text_len);
 
 	/* printf ("--- after split\n");
-	printf ("left\n");
-	debug_spell_errors (HTML_TEXT (self)->spell_errors);
-	printf ("right\n");
-	debug_spell_errors (HTML_TEXT (dup)->spell_errors);
-	printf ("---\n"); */
+	   printf ("left\n");
+	   debug_spell_errors (HTML_TEXT (self)->spell_errors);
+	   printf ("right\n");
+	   debug_spell_errors (HTML_TEXT (dup)->spell_errors);
+	   printf ("---\n");
+	*/
 
 	*left  = g_list_prepend (*left, self);
 	*right = g_list_prepend (*right, dup);
@@ -768,11 +806,10 @@ update_asc_dsc (HTMLPainter *painter, PangoItem *item, gint *asc, gint *dsc)
 	pango_font_metrics_unref (pfm);
 }
 
-PangoAttrList *
-html_text_get_attr_list (HTMLText *text, gint start_index, gint end_index)
+void
+html_text_get_attr_list_list (PangoAttrList *get_attrs, PangoAttrList *attr_list, gint start_index, gint end_index)
 {
-	PangoAttrIterator *iter = pango_attr_list_get_iterator (text->attr_list);
-	PangoAttrList *attrs = pango_attr_list_new ();
+	PangoAttrIterator *iter = pango_attr_list_get_iterator (attr_list);
 
 	if (iter) {
 		do {
@@ -797,12 +834,22 @@ html_text_get_attr_list (HTMLText *text, gint start_index, gint end_index)
 						attr->end_index -= start_index;
 
 					c->data = NULL;
-					pango_attr_list_insert (attrs, attr);
+					pango_attr_list_insert (get_attrs, attr);
 				}
 				g_slist_free (l);
 			}
 		} while (pango_attr_iterator_next (iter));
 	}
+}
+
+PangoAttrList *
+html_text_get_attr_list (HTMLText *text, gint start_index, gint end_index)
+{
+	PangoAttrList *attrs = pango_attr_list_new ();
+
+	html_text_get_attr_list_list (attrs, text->attr_list, start_index, end_index);
+	if (text->extra_attr_list)
+		html_text_get_attr_list_list (attrs, text->extra_attr_list, start_index, end_index);
 
 	return attrs;
 }
@@ -1046,8 +1093,16 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 			attr->start_index = 0;
 			attr->end_index = text->text_bytes;
 			pango_attr_list_insert (attrs, attr);
+			if (painter->font_manager.fix_size != painter->font_manager.var_size) {
+				attr = pango_attr_size_new (painter->font_manager.fix_size);
+				attr->start_index = 0;
+				attr->end_index = text->text_bytes;
+				pango_attr_list_insert (attrs, attr);
+			}
 		} else
 			attrs = pango_attr_list_copy (text->attr_list);
+		if (text->extra_attr_list)
+			pango_attr_list_splice (attrs, text->extra_attr_list, 0, text->text_len);
 
 		if (text->select_length) {
 			gchar *end;
@@ -1137,8 +1192,8 @@ html_text_tail_white_space (HTMLText *text, HTMLPainter *painter, gint offset, g
 		s = g_utf8_prev_char (s);
 		if (pi->entries [ii].attrs [io].is_white) {
 			if (HTML_IS_GDK_PAINTER (painter) || HTML_IS_PLAIN_PAINTER (painter)) {
-				if (*s == '\t') {
-					gint skip = 8, co = offset;
+				if (*s == '\t' && offset > 1) {
+					gint skip = 8, co = offset - 1;
 
 					do {
 						s = g_utf8_prev_char (s);
@@ -1274,14 +1329,15 @@ save_open_attrs (HTMLEngineSaveState *state, GSList *attrs)
 		case PANGO_ATTR_STRIKETHROUGH:
 			tag = "<S>";
 			break;
-		case PANGO_ATTR_SIZE: {
-			HTMLPangoAttrFontSize *size = (HTMLPangoAttrFontSize *) attr;
-			if ((size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != GTK_HTML_FONT_STYLE_SIZE_3) {
-				tag = g_strdup_printf ("<FONT SIZE=\"%d\">", size->style & GTK_HTML_FONT_STYLE_SIZE_MASK);
-				free_tag = TRUE;
+		case PANGO_ATTR_SIZE:
+			if (attr->klass == &html_pango_attr_font_size_klass) {
+				HTMLPangoAttrFontSize *size = (HTMLPangoAttrFontSize *) attr;
+				if ((size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != GTK_HTML_FONT_STYLE_SIZE_3) {
+					tag = g_strdup_printf ("<FONT SIZE=\"%d\">", size->style & GTK_HTML_FONT_STYLE_SIZE_MASK);
+					free_tag = TRUE;
+				}
 			}
-		}
-		break;
+			break;
 		case PANGO_ATTR_FAMILY:
 			/* TODO */
 			break;
@@ -1438,7 +1494,7 @@ save (HTMLObject *self, HTMLEngineSaveState *state)
 
 			attrs = pango_attr_iterator_get_attrs (iter);
 			pango_attr_iterator_range (iter, &start_index, &end_index);
-			if (end_index == G_MAXINT)
+			if (end_index > text->text_bytes)
 				end_index = text->text_bytes;
 
 			if (l && !link_started) {
@@ -1615,6 +1671,7 @@ html_text_convert_nbsp (HTMLText *text, gboolean free_text)
 	if (is_convert_nbsp_needed (text->text, &delta)) {
 		to_free    = text->text;
 		text->text = g_malloc (strlen (to_free) + delta + 1);
+		text->text_bytes += delta;
 		convert_nbsp (text->text, to_free);
 		if (free_text)
 			g_free (to_free);
@@ -1629,7 +1686,8 @@ move_spell_errors (GList *spell_errors, guint offset, gint delta)
 { 
 	SpellError *se; 
 
-	if (!delta) return;
+	if (!delta)
+		return;
 
 	while (spell_errors) { 
 		se = (SpellError *) spell_errors->data; 
@@ -1766,6 +1824,10 @@ destroy (HTMLObject *obj)
 	pango_info_destroy (text);
 	pango_attr_list_unref (text->attr_list);
 	text->attr_list = NULL;
+	if (text->extra_attr_list) {
+		pango_attr_list_unref (text->extra_attr_list);
+		text->extra_attr_list = NULL;
+	}
 	free_links (text->links);
 	text->links = NULL;
 
@@ -2066,6 +2128,7 @@ html_text_init (HTMLText *text,
 	text->select_length = 0;
 	text->pi            = NULL;
 	text->attr_list     = pango_attr_list_new ();
+	text->extra_attr_list = NULL;
 	text->links         = NULL;
 
 	html_color_ref (color);
@@ -2701,25 +2764,25 @@ html_pango_attr_font_size_calc (HTMLPangoAttrFontSize *attr, HTMLEngine *e)
 {
 	gint size, base_size, real_size;
 
-	base_size = e->painter->font_manager.var_size;
+	base_size = (attr->style & GTK_HTML_FONT_STYLE_FIXED) ? e->painter->font_manager.fix_size : e->painter->font_manager.var_size;
 	size = (attr->style & GTK_HTML_FONT_STYLE_SIZE_MASK) - GTK_HTML_FONT_STYLE_SIZE_3;
 	real_size = e->painter->font_manager.magnification * ((gdouble) base_size + (size > 0 ? (1 << size) : size) * base_size/8.0);
 
 	attr->attr_int.value = real_size;
 }
 
+static const PangoAttrClass html_pango_attr_font_size_klass = {
+	PANGO_ATTR_SIZE,
+	html_pango_attr_font_size_copy,
+	html_pango_attr_font_size_destroy,
+	html_pango_attr_font_size_equal
+};
+
 PangoAttribute *
 html_pango_attr_font_size_new (GtkHTMLFontStyle style)
 {
-	static const PangoAttrClass klass = {
-		PANGO_ATTR_SIZE,
-		html_pango_attr_font_size_copy,
-		html_pango_attr_font_size_destroy,
-		html_pango_attr_font_size_equal
-	};
-
 	HTMLPangoAttrFontSize *result = g_new (HTMLPangoAttrFontSize, 1);
-	result->attr_int.attr.klass = &klass;
+	result->attr_int.attr.klass = &html_pango_attr_font_size_klass;
 	result->style = style;
 
 	return (PangoAttribute *) result;
@@ -2865,7 +2928,7 @@ html_text_set_style_in_range (HTMLText *text, GtkHTMLFontStyle style, HTMLEngine
 
 	/* size */
 	if (style & GTK_HTML_FONT_STYLE_SIZE_MASK) {
-		attr = html_pango_attr_font_size_new (style & GTK_HTML_FONT_STYLE_SIZE_MASK);
+		attr = html_pango_attr_font_size_new (style);
 		html_pango_attr_font_size_calc ((HTMLPangoAttrFontSize *) attr, e);
 		attr->start_index = start_index;
 		attr->end_index = end_index;
