@@ -25,8 +25,8 @@
 /* The HTML Tokenizer */
 #include <config.h>
 #include <ctype.h>
+#include <string.h>
 #include <gnome.h>
-#include <gal/unicode/gunicode.h>
 #include "htmltokenizer.h"
 #include "htmlentity.h"
 
@@ -39,6 +39,7 @@ enum {
 static guint html_tokenizer_signals[HTML_TOKENIZER_LAST_SIGNAL] = { 0 };
 
 #define TOKEN_BUFFER_SIZE (1 << 10)
+#define INVALID_CHARACTER_MARKER '?'
 
 typedef struct _HTMLBlockingToken HTMLBlockingToken;
 typedef struct _HTMLTokenBuffer   HTMLTokenBuffer;
@@ -139,7 +140,7 @@ static void           html_tokenizer_append_token_buffer (HTMLTokenizer *t,
 							  gint min_size);
 
 /* default implementations of tokenization functions */
-static void     html_tokenizer_destructor          (GtkObject *);
+static void     html_tokenizer_finalize             (GObject *);
 static void     html_tokenizer_real_begin           (HTMLTokenizer *, gchar *content_type);
 static void     html_tokenizer_real_write           (HTMLTokenizer *, const gchar *str, size_t size);
 static void     html_tokenizer_real_end             (HTMLTokenizer *);
@@ -159,36 +160,36 @@ static void               html_tokenizer_tokenize_one_char   (HTMLTokenizer  *t,
 
 static void               add_unichar(HTMLTokenizer *t, gunichar wc);
 
-static GtkObjectClass *parent_class = NULL;
+static GObjectClass *parent_class = NULL;
 
 static void
 html_tokenizer_class_init (HTMLTokenizerClass *klass)
 {
-	GtkObjectClass *object_class = (GtkObjectClass *) klass;
+	GObjectClass *object_class = (GObjectClass *) klass;
 
-	parent_class = gtk_type_class (GTK_TYPE_OBJECT);
+	parent_class = g_type_class_ref (G_TYPE_OBJECT);
 
 	html_tokenizer_signals[HTML_TOKENIZER_BEGIN_SIGNAL] =
-		gtk_signal_new ("begin",
-				GTK_RUN_LAST,
-				object_class->type,
-				GTK_SIGNAL_OFFSET (HTMLTokenizerClass, begin),
-				gtk_marshal_NONE__POINTER,
-				GTK_TYPE_NONE,
-				1, GTK_TYPE_POINTER);
+		g_signal_new ("begin",
+			      G_TYPE_FROM_CLASS (klass),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (HTMLTokenizerClass, begin),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__POINTER,
+			      G_TYPE_NONE,
+			      1, G_TYPE_POINTER);
 
-	html_tokenizer_signals[HTML_TOKENIZER_END_SIGNAL] =
-		gtk_signal_new ("end",
-				GTK_RUN_LAST,
-				object_class->type,
-				GTK_SIGNAL_OFFSET (HTMLTokenizerClass, end),
-				gtk_marshal_NONE__NONE,
-				GTK_TYPE_NONE,
-				0);
+	html_tokenizer_signals [HTML_TOKENIZER_END_SIGNAL] =
+		g_signal_new ("end",
+			      G_TYPE_FROM_CLASS (klass),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (HTMLTokenizerClass, end),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE,
+			      0);
 
-	gtk_object_class_add_signals (object_class, html_tokenizer_signals, HTML_TOKENIZER_LAST_SIGNAL);
-
-	object_class->destroy = html_tokenizer_destructor;
+	object_class->finalize = html_tokenizer_finalize;
 
 	klass->begin      = html_tokenizer_real_begin;
 	klass->end        = html_tokenizer_real_end;
@@ -248,7 +249,7 @@ html_tokenizer_init (HTMLTokenizer *t)
 }
 
 static void
-html_tokenizer_destructor (GtkObject *obj)
+html_tokenizer_finalize (GObject *obj)
 {
 	HTMLTokenizer *t = HTML_TOKENIZER (obj);
 
@@ -264,16 +265,18 @@ html_tokenizer_get_type (void)
 	static GtkType html_tokenizer_type = 0;
 
 	if (!html_tokenizer_type) {
-		static const GtkTypeInfo html_tokenizer_info = {
-			"HTMLTokenizer",
-			sizeof (HTMLTokenizer),
+		static const GTypeInfo html_tokenizer_info = {
 			sizeof (HTMLTokenizerClass),
-			(GtkClassInitFunc) html_tokenizer_class_init,
-			(GtkObjectInitFunc) html_tokenizer_init,
-			NULL, NULL,
-			(GtkClassInitFunc) NULL
+			NULL,
+			NULL,
+			(GClassInitFunc) html_tokenizer_class_init,
+			NULL,
+			NULL,
+			sizeof (HTMLTokenizer),
+			1,
+			(GInstanceInitFunc) html_tokenizer_init,
 		};
-		html_tokenizer_type = gtk_type_unique (GTK_TYPE_OBJECT, &html_tokenizer_info);
+		html_tokenizer_type = g_type_register_static (G_TYPE_OBJECT, "HTMLTokenizer", &html_tokenizer_info, 0);
 	}
 
 	return html_tokenizer_type;
@@ -319,7 +322,7 @@ html_token_buffer_append_token (HTMLTokenBuffer * buf, const gchar *token, gint
 HTMLTokenizer *
 html_tokenizer_new (void)
 {
-	return (HTMLTokenizer *) gtk_type_new (HTML_TYPE_TOKENIZER);
+	return (HTMLTokenizer *) g_object_new (HTML_TYPE_TOKENIZER, NULL);
 }
 
 void
@@ -327,7 +330,7 @@ html_tokenizer_destroy (HTMLTokenizer *t)
 {
 	g_return_if_fail (t && HTML_IS_TOKENIZER (t));
 	
-	gtk_object_unref (GTK_OBJECT (t));
+	g_object_unref (G_OBJECT (t));
 }
 
 static gchar *
@@ -705,13 +708,8 @@ in_script_or_style (HTMLTokenizer *t, const gchar **src)
 	struct _HTMLTokenizerPrivate *p = t->priv;
 	
 	/* Allocate memory to store the script or style */
-	if (p->scriptCodeSize + 11 > p->scriptCodeMaxSize) {
-		gchar *newbuf = g_malloc (p->scriptCodeSize + 1024);
-		memcpy (newbuf, p->scriptCode, p->scriptCodeSize);
-		g_free (p->scriptCode);
-
-		p->scriptCodeMaxSize += 1024;
-	}
+	if (p->scriptCodeSize + 11 > p->scriptCodeMaxSize)
+		p->scriptCode = g_realloc (p->scriptCode, p->scriptCodeMaxSize += 1024);
 			
 	if ((**src == '>' ) && ( p->searchFor [p->searchCount] == '>')) {
 		(*src)++;
@@ -762,11 +760,10 @@ add_unichar (HTMLTokenizer *t, gunichar wc)
 
 	p->utf8_length = 0;
 
-	if (wc != 0) {
+	if (wc != '\0') {
 		p->dest += g_unichar_to_utf8 (wc, p->dest);
 		*(p->dest) = 0;
 	}
-	/* g_assert (g_utf8_validate (p->buffer, p->dest - p->buffer, NULL)); */
 }
 
 static void
@@ -781,7 +778,7 @@ add_byte (HTMLTokenizer *t, const gchar **src)
 
 		wc = g_utf8_get_char_validated ((const gchar *)p->utf8_buffer, p->utf8_length);
 		if (wc == -1 || p->utf8_length >= (sizeof(p->utf8_buffer)/sizeof(p->utf8_buffer[0]))) {
-			add_unichar (t, '?');
+			add_unichar (t, INVALID_CHARACTER_MARKER);
 			(*src)++;
 			return;
 		} else if (wc == -2) {
@@ -809,6 +806,22 @@ flush_entity (HTMLTokenizer *t)
 	}
 }
 
+static gboolean
+add_unichar_validated (HTMLTokenizer *t, gunichar uc)
+{
+	char tmp[8];
+
+	tmp [g_unichar_to_utf8 (uc, tmp)] = '\0';
+
+	if (g_utf8_validate (tmp, -1, NULL)) {
+		add_unichar (t, uc);
+		return TRUE;
+	} 
+		
+	g_warning ("invalid character value: x%xd", uc);
+	return FALSE;
+}
+
 static void
 in_entity (HTMLTokenizer *t, const gchar **src)
 {
@@ -838,6 +851,7 @@ in_entity (HTMLTokenizer *t, const gchar **src)
 		    (p->searchBuffer[3] == 'x')) {
 			/* &x12AB */
 			p->searchBuffer [p->searchCount + 1] = '\0';
+			
 			entityValue = strtoul (&(p->searchBuffer [4]),
 					       NULL, 16);
 			p->charEntity = FALSE;
@@ -858,7 +872,7 @@ in_entity (HTMLTokenizer *t, const gchar **src)
 				
 	}
 
-	if (p->searchCount > 9) {
+	if (p->searchCount > 13) {
 		/* Ignore this sequence since it's too long */
 		p->charEntity = FALSE;
 		flush_entity (t);
@@ -875,12 +889,14 @@ in_entity (HTMLTokenizer *t, const gchar **src)
 		 * and attributes. 
 		 */
 		if (entityValue) {
-			/* Insert plain char */
-			add_unichar (t, entityValue);
+			if (entityValue != TAG_ESCAPE) 
+				/* make sure the entity value is a valid character value */
+				if (!add_unichar_validated (t, entityValue))
+					add_unichar (t, INVALID_CHARACTER_MARKER);
+
 			if (**src == ';')
 				(*src)++;
-		}
-		else {
+		} else {
 			/* Ignore the sequence, just add it as plaintext */
 			flush_entity (t);
 		}
@@ -1029,10 +1045,7 @@ end_tag (HTMLTokenizer *t, const gchar **src)
 	else if (strncmp (p->buffer + 2, "/select", 7) == 0) {
 		p->select = FALSE;
 	}
-	else if (strncmp (p->buffer + 2, "cell", 4) == 0) {
-		g_warning ("<cell> tag not supported");
-	}
-	else if (strncmp (p->buffer + 2, "table", 5) == 0) {
+	else if (strncmp (p->buffer + 2, "tablesdkl", 9) == 0) {
 		html_tokenizer_blocking_push (t, Table);
 	}
 	else {
@@ -1260,7 +1273,7 @@ html_tokenizer_blocking_get_name (HTMLTokenizer *t)
 {
 	switch (GPOINTER_TO_INT (t->priv->blocking->data)) {
 	case Table:
-		return "</table";
+		return "</tabledkdk";
 	}
 	
 	return "";
@@ -1300,9 +1313,7 @@ html_tokenizer_begin (HTMLTokenizer *t, gchar *content_type)
 {
 	g_return_if_fail (t && HTML_IS_TOKENIZER (t));
 
-	gtk_signal_emit (GTK_OBJECT (t),
-			 html_tokenizer_signals[HTML_TOKENIZER_BEGIN_SIGNAL],
-			 content_type);
+	g_signal_emit (t, html_tokenizer_signals [HTML_TOKENIZER_BEGIN_SIGNAL], 0, content_type);
 }
 
 void
@@ -1310,8 +1321,7 @@ html_tokenizer_end (HTMLTokenizer *t)
 {
 	g_return_if_fail (t && HTML_IS_TOKENIZER (t));
 
-	gtk_signal_emit (GTK_OBJECT (t),
-			 html_tokenizer_signals[HTML_TOKENIZER_END_SIGNAL]);
+	g_signal_emit (t, html_tokenizer_signals[HTML_TOKENIZER_END_SIGNAL], 0);
 }
 
 void
@@ -1320,8 +1330,8 @@ html_tokenizer_write (HTMLTokenizer *t, const gchar *str, size_t size)
 	HTMLTokenizerClass *klass;
 
 	g_return_if_fail (t && HTML_IS_TOKENIZER (t));
-	klass = HTML_TOKENIZER_CLASS (GTK_OBJECT (t)->klass);
-	
+	klass = HTML_TOKENIZER_CLASS (G_OBJECT_GET_CLASS (t));
+
 	if (klass->write)
 		klass->write (t, str, size);
 	else
@@ -1335,8 +1345,8 @@ html_tokenizer_peek_token (HTMLTokenizer *t)
 
 	g_return_val_if_fail (t && HTML_IS_TOKENIZER (t), NULL);
 
-	klass = HTML_TOKENIZER_CLASS (GTK_OBJECT (t)->klass);
-	
+	klass = HTML_TOKENIZER_CLASS (G_OBJECT_GET_CLASS (t));
+
 	if (klass->peek_token)
 		return klass->peek_token (t);
 	
@@ -1352,8 +1362,8 @@ html_tokenizer_next_token (HTMLTokenizer *t)
 
 	g_return_val_if_fail (t && HTML_IS_TOKENIZER (t), NULL);
 
-	klass = HTML_TOKENIZER_CLASS (GTK_OBJECT (t)->klass);
-	
+	klass = HTML_TOKENIZER_CLASS (G_OBJECT_GET_CLASS (t));
+
 	if (klass->next_token)
 		return klass->next_token (t);
 
@@ -1368,8 +1378,8 @@ html_tokenizer_has_more_tokens (HTMLTokenizer *t)
 
 	g_return_val_if_fail (t && HTML_IS_TOKENIZER (t), FALSE);
 
-	klass = HTML_TOKENIZER_CLASS (GTK_OBJECT (t)->klass);
-	
+	klass = HTML_TOKENIZER_CLASS (G_OBJECT_GET_CLASS (t));
+
 	if (klass->has_more) {
 		return klass->has_more (t);
 	}
@@ -1388,8 +1398,8 @@ html_tokenizer_clone (HTMLTokenizer *t)
 		return NULL;
 	g_return_val_if_fail (HTML_IS_TOKENIZER (t), NULL);
 
-	klass = HTML_TOKENIZER_CLASS (GTK_OBJECT (t)->klass);
-	
+	klass = HTML_TOKENIZER_CLASS (G_OBJECT_GET_CLASS (t));
+
 	if (klass->clone)
 		return klass->clone (t);
 	
